"""Text parser plugin for syslog log files.

Also see:
  https://docs.rsyslog.com/doc//configuration/templates.html
"""

import re

from dfdatetime import time_elements as dfdatetime_time_elements

import pyparsing

from plaso.containers import events
from plaso.lib import dateless_helper
from plaso.lib import errors
from plaso.parsers import logger
from plaso.parsers import text_parser
from plaso.parsers.text_plugins import interface


class SyslogCommentEventData(events.EventData):
  """Syslog comment event data.

  Attributes:
    body (str): message body.
    last_written_time (dfdatetime.DateTimeValues): entry last written date and
        time.
  """

  DATA_TYPE = 'syslog:comment'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.body = None
    self.last_written_time = None


class SyslogLineEventData(events.EventData):
  """Syslog line event data.

  Attributes:
    body (str): message body.
    facility (str): facility.
    hostname (str): hostname of the reporter.
    last_written_time (dfdatetime.DateTimeValues): entry last written date and
        time.
    message_identifier (str): message identifier.
    pid (str): process identifier of the reporter.
    reporter (str): reporter.
    severity (str): severity.
  """

  DATA_TYPE = 'syslog:line'

  def __init__(self, data_type=DATA_TYPE):
    """Initializes an event data attribute container.

    Args:
      data_type (Optional[str]): event data type indicator.
    """
    super().__init__(data_type=data_type)
    self.body = None
    self.facility = None
    self.hostname = None
    self.last_written_time = None
    self.message_identifier = None
    self.pid = None
    self.reporter = None
    self.severity = None


class SyslogCronTaskRunEventData(SyslogLineEventData):
  """Syslog cron task run event data.

  Attributes:
    command (str): command executed.
    last_written_time (dfdatetime.DateTimeValues): entry last written date and
        time.
    username (str): name of user the command was executed.
  """

  DATA_TYPE = 'syslog:cron:task_run'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.command = None
    self.last_written_time = None
    self.username = None


class SyslogSSHEventData(SyslogLineEventData):
  """SSH event data.

  Attributes:
    authentication_method (str): authentication method.
    fingerprint (str): fingerprint.
    ip_address (str): IP address.
    last_written_time (dfdatetime.DateTimeValues): entry last written date and
        time.
    port (str): port.
    protocol (str): protocol.
    username (str): name of user the command was executed.
  """

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.authentication_method = None
    self.fingerprint = None
    self.ip_address = None
    self.last_written_time = None
    self.port = None
    self.protocol = None
    self.username = None


# TODO: merge separate SyslogSSHEventData classes.
class SyslogSSHLoginEventData(SyslogSSHEventData):
  """SSH login event data."""

  DATA_TYPE = 'syslog:ssh:login'


class SyslogSSHFailedConnectionEventData(SyslogSSHEventData):
  """SSH failed connection event data."""

  DATA_TYPE = 'syslog:ssh:failed_connection'


class SyslogSSHOpenedConnectionEventData(SyslogSSHEventData):
  """SSH opened connection event data."""

  DATA_TYPE = 'syslog:ssh:opened_connection'


class BaseSyslogTextPlugin(interface.TextPlugin):
  """Shared functionality for syslog log file text parser plugins."""

  # pylint: disable=abstract-method

  _CRON_USERNAME = (
      pyparsing.Literal('(') +
      pyparsing.Word(pyparsing.alphanums).set_results_name('username') +
      pyparsing.Literal(')'))

  _CRON_COMMAND_END = pyparsing.Literal(')') + pyparsing.StringEnd()

  _CRON_COMMAND = (
      pyparsing.Literal('CMD') + pyparsing.Literal('(') +
      pyparsing.SkipTo(_CRON_COMMAND_END).set_results_name('command') +
      _CRON_COMMAND_END)

  _CRON_TASK_RUN = _CRON_USERNAME + _CRON_COMMAND + pyparsing.StringEnd()

  _CRON_MESSAGE = pyparsing.Group(_CRON_TASK_RUN).set_results_name('task_run')

  _SSHD_AUTHENTICATION_METHOD = (
      pyparsing.Keyword('password') | pyparsing.Keyword('publickey'))

  _SSHD_FINGER_PRINT = pyparsing.Combine(
      pyparsing.Literal('RSA ') +
      pyparsing.Word(':' + pyparsing.hexnums)).set_results_name('fingerprint')

  _SSH_USERNAME = pyparsing.Word(pyparsing.alphanums).set_results_name(
      'username')

  _SSH_IP_ADDRESS = (
      pyparsing.pyparsing_common.ipv4_address |
      pyparsing.pyparsing_common.ipv6_address)

  _SSH_PORT = pyparsing.Word(pyparsing.nums, max=5).set_results_name('port')

  _SSHD_FAILED_CONNECTION = (
      pyparsing.Literal('Failed') +
      _SSHD_AUTHENTICATION_METHOD.set_results_name('authentication_method') +
      pyparsing.Literal('for') + _SSH_USERNAME +
      pyparsing.Literal('from') +
      _SSH_IP_ADDRESS.set_results_name('ip_address') +
      pyparsing.Literal('port') + _SSH_PORT +
      pyparsing.StringEnd())

  _SSHD_LOGIN = (
      pyparsing.Literal('Accepted') +
      _SSHD_AUTHENTICATION_METHOD.set_results_name('authentication_method') +
      pyparsing.Literal('for') + _SSH_USERNAME +
      pyparsing.Literal('from') +
      _SSH_IP_ADDRESS.set_results_name('ip_address') +
      pyparsing.Literal('port') + _SSH_PORT +
      pyparsing.Literal('ssh2').set_results_name('protocol') +
      pyparsing.Optional(pyparsing.Literal(':') + _SSHD_FINGER_PRINT) +
      pyparsing.StringEnd())

  _SSHD_OPENED_CONNECTION = (
      pyparsing.Literal('Connection from') +
      _SSH_IP_ADDRESS.set_results_name('ip_address') +
      pyparsing.Literal('port') + _SSH_PORT +
      pyparsing.StringEnd())

  _SSHD_MESSAGE = (
      pyparsing.Group(_SSHD_FAILED_CONNECTION).set_results_name(
          'failed_connection') ^
      pyparsing.Group(_SSHD_LOGIN).set_results_name('login') ^
      pyparsing.Group(_SSHD_OPENED_CONNECTION).set_results_name(
          'opened_connection'))

  def _ParseCronMessageBody(self, body):
    """Parses a cron syslog message body.

    Args:
      body (str): syslog message body.

    Returns:
      SyslogCronTaskRunEventData: event data or None if not available.
    """
    try:
      structure = self._CRON_MESSAGE.parse_string(body)
    except pyparsing.ParseException as exception:
      logger.debug(
          f'Unable to parse cron message body with error: {exception!s}')
      return None

    keys = list(structure.keys())
    if len(keys) != 1:
      return None

    key = keys[0]
    structure = structure[0]

    if key != 'task_run':
      return None

    event_data = SyslogCronTaskRunEventData()
    event_data.body = structure.get('body')
    event_data.command = structure.get('command')
    event_data.hostname = structure.get('hostname')
    event_data.pid = structure.get('pid')
    event_data.reporter = structure.get('reporter')
    event_data.severity = structure.get('severity')
    event_data.username = structure.get('username')

    return event_data

  def _ParseSshdMessageBody(self, body):
    """Parses a sshd syslog message body.

    Args:
      body (str): syslog message body.

    Returns:
      SyslogCronTaskRunEventData: event data or None if not available.
    """
    try:
      structure = self._SSHD_MESSAGE.parse_string(body)
    except pyparsing.ParseException as exception:
      logger.debug(
          f'Unable to parse sshd message body with error: {exception!s}')
      return None

    keys = list(structure.keys())
    if len(keys) != 1:
      return None

    key = keys[0]
    structure = structure[0]

    if key == 'failed_connection':
      event_data = SyslogSSHFailedConnectionEventData()
    elif key == 'login':
      event_data = SyslogSSHLoginEventData()
    elif key == 'opened_connection':
      event_data = SyslogSSHOpenedConnectionEventData()
    else:
      return None

    event_data.authentication_method = structure.get(
        'authentication_method', None)
    event_data.body = structure.get('body')
    event_data.fingerprint = structure.get('fingerprint')
    event_data.hostname = structure.get('hostname')
    event_data.ip_address = structure.get('ip_address')
    event_data.pid = structure.get('pid')
    event_data.protocol = structure.get('protocol')
    event_data.port = structure.get('port')
    event_data.reporter = structure.get('reporter')
    event_data.severity = structure.get('severity')
    event_data.username = structure.get('username')

    return event_data


class SyslogTextPlugin(BaseSyslogTextPlugin):
  """Text parser plugin for syslog log files."""

  NAME = 'syslog'
  DATA_FORMAT = 'System log (syslog) file'

  ENCODING = 'utf-8'

  # The reporter and facility fields can contain any printable character, but
  # to allow for processing of syslog formats that delimit the reporter and
  # facility with printable characters, we remove certain common delimiters
  # from the set of printable characters.

  _REPORTER_CHARACTERS = ''.join(
      [c for c in pyparsing.printables if c not in [':', '[', '<']])

  _FACILITY_CHARACTERS = ''.join(
      [c for c in pyparsing.printables if c not in [':', '>']])

  # Note that the following values are sorted in-order of their corresponding
  # syslog protocol 23 priority value.
  _SYSLOG_SEVERITY = [
      'EMERG', 'ALERT', 'CRIT', 'ERR', 'WARNING', 'NOTICE', 'INFO', 'DEBUG']

  # According to section 6.2.1 of
  # https://datatracker.ietf.org/doc/html/draft-ietf-syslog-protocol-23
  #  0             kernel messages
  #  1             user-level messages
  #  2             mail system
  #  3             system daemons
  #  4             security/authorization messages
  #  5             messages generated internally by syslogd
  #  6             line printer subsystem
  #  7             network news subsystem
  #  8             UUCP subsystem
  #  9             clock daemon
  # 10             security/authorization messages
  # 11             FTP daemon
  # 12             NTP subsystem
  # 13             log audit
  # 14             log alert
  # 15             clock daemon (note 2)
  # 16             local use 0  (local0)
  # 17             local use 1  (local1)
  # 18             local use 2  (local2)
  # 19             local use 3  (local3)
  # 20             local use 4  (local4)
  # 21             local use 5  (local5)
  # 22             local use 6  (local6)
  # 23             local use 7  (local7)
  _SYSLOG_FACILITY = [
      'kernel message',
      'user-level message',
      'mail system',
      'system daemons',
      'security/authorization messages',
      'messages generated internally by syslogd',
      'line printer subsystem',
      'network news subsystem',
      'UUCP subsystem',
      'clock daemon',
      'security/authorization messages',
      'FTP daemon',
      'NTP subsystem',
      'log audit',
      'log alert',
      'clock daemon',
      'local use 0',
      'local use 1',
      'local use 2',
      'local use 3',
      'local use 4',
      'local use 5',
      'local use 6',
      'local use 7']

  # TODO: change pattern to allow only spaces as a field separator.
  _BODY_PATTERN = (
      r'.*?(?=($|\n\w{3}\s+\d{1,2}\s\d{2}:\d{2}:\d{2})|'
      r'($|\n\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\.\d{6}[\+|-]\d{2}:\d{2}\s)|'
      r'($|\n<\d{1,3}>1\s\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\.\d{6}[\+|-]\d{2}'
      r':\d{2}\s))')

  _ONE_OR_TWO_DIGITS = pyparsing.Word(pyparsing.nums, max=2).set_parse_action(
      lambda tokens: int(tokens[0], 10))

  _TWO_DIGITS = pyparsing.Word(pyparsing.nums, exact=2).set_parse_action(
      lambda tokens: int(tokens[0], 10))

  _FOUR_DIGITS = pyparsing.Word(pyparsing.nums, exact=4).set_parse_action(
      lambda tokens: int(tokens[0], 10))

  _SIX_DIGITS = pyparsing.Word(pyparsing.nums, exact=6).set_parse_action(
      lambda tokens: int(tokens[0], 10))

  _DATE_TIME_RFC3339 = (
      _FOUR_DIGITS + pyparsing.Suppress('-') +
      _TWO_DIGITS + pyparsing.Suppress('-') +
      _TWO_DIGITS + pyparsing.Suppress('T') +
      _TWO_DIGITS + pyparsing.Suppress(':') +
      _TWO_DIGITS + pyparsing.Suppress(':') +
      _TWO_DIGITS + pyparsing.Suppress('.') +
      _SIX_DIGITS + pyparsing.Word('+-', exact=1) +
      _TWO_DIGITS + pyparsing.Optional(
          pyparsing.Suppress(':') + _TWO_DIGITS))

  _PROCESS_IDENTIFIER = pyparsing.Word(pyparsing.nums, max=5).set_parse_action(
      lambda tokens: int(tokens[0], 10))

  _REPORTER = pyparsing.Word(_REPORTER_CHARACTERS)

  _END_OF_LINE = pyparsing.Suppress(pyparsing.LineEnd())

  # The ChromeOS syslog messages are of a format beginning with an
  # ISO 8601 combined date and time expression with a time zone offset:
  #   2016-10-25T12:37:23.297265-07:00
  #
  # This will then be followed by the SYSLOG Severity which will be one of:
  #   EMERG,ALERT,CRIT,ERR,WARNING,NOTICE,INFO,DEBUG
  #
  # 2016-10-25T12:37:23.297265-07:00 INFO

  _CHROMEOS_SYSLOG_LINE_BODY = (
      pyparsing.one_of(_SYSLOG_SEVERITY).set_results_name('severity') +
      _REPORTER.set_results_name('reporter') +
      pyparsing.Optional(pyparsing.Suppress(':')) +
      pyparsing.Optional(
          pyparsing.Suppress('[') +
          _PROCESS_IDENTIFIER.set_results_name('pid') +
          pyparsing.Suppress(']')))

  # The rsyslog file format (RSYSLOG_FileFormat) consists of:
  # %TIMESTAMP% %HOSTNAME% %syslogtag%%msg%
  #
  # Where %TIMESTAMP% is in RFC-3339 date time format e.g.
  # 2020-05-31T00:00:45.698463+00:00

  _RSYSLOG_LINE_BODY = (
      pyparsing.Word(pyparsing.printables).set_results_name('hostname') +
      _REPORTER.set_results_name('reporter') +
      pyparsing.Optional(
          pyparsing.Suppress('[') +
          _PROCESS_IDENTIFIER.set_results_name('pid') +
          pyparsing.Suppress(']')) +
      pyparsing.Optional(
          pyparsing.Suppress('<') +
          pyparsing.Word(_FACILITY_CHARACTERS).set_results_name('facility') +
          pyparsing.Suppress('>')))

  _LOG_LINE = (
      _DATE_TIME_RFC3339.set_results_name('date_time') +
      (_CHROMEOS_SYSLOG_LINE_BODY ^ _RSYSLOG_LINE_BODY) +
      pyparsing.Optional(pyparsing.Suppress(':')) +
      pyparsing.Regex(_BODY_PATTERN, re.DOTALL).set_results_name('body') +
      _END_OF_LINE)

  # The rsyslog protocol 23 format (RSYSLOG_SyslogProtocol23Format)
  # consists of:
  # %PRI%1 %TIMESTAMP% %HOSTNAME% %APP-NAME% %PROCID% %MSGID% %STRUCTURED-DATA%
  #   %msg%
  #
  # Where %TIMESTAMP% is in RFC-3339 date time format e.g.
  # 2020-05-31T00:00:45.698463+00:00

  # TODO: Add proper support for %STRUCTURED-DATA%:
  # https://datatracker.ietf.org/doc/html/draft-ietf-syslog-protocol-23#section-6.3
  _RSYSLOG_PROTOCOL_23_LINE = (
      pyparsing.Suppress('<') +
      _ONE_OR_TWO_DIGITS.set_results_name('priority') +
      pyparsing.Suppress('>') + pyparsing.Suppress(
          pyparsing.Word(pyparsing.nums, max=1)) +
      _DATE_TIME_RFC3339.set_results_name('date_time') +
      pyparsing.Word(pyparsing.printables).set_results_name('hostname') +
      _REPORTER.set_results_name('reporter') +
      pyparsing.Or([pyparsing.Suppress('-'),
                    _PROCESS_IDENTIFIER.set_results_name('pid')]) +
      pyparsing.Word(pyparsing.printables).set_results_name(
          'message_identifier') +
      pyparsing.Word(pyparsing.printables).set_results_name('structured_data') +
      pyparsing.Regex(_BODY_PATTERN, re.DOTALL).set_results_name('body') +
      _END_OF_LINE)

  _LINE_STRUCTURES = [
      ('log_line', _LOG_LINE),
      ('rsyslog_protocol_23_line', _RSYSLOG_PROTOCOL_23_LINE)]

  VERIFICATION_GRAMMAR = _LOG_LINE ^ _RSYSLOG_PROTOCOL_23_LINE

  def _ParseRecord(self, parser_mediator, key, structure):
    """Parses a pyparsing structure.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      key (str): name of the parsed structure.
      structure (pyparsing.ParseResults): tokens from a parsed log line.

    Raises:
      ParseError: if the structure cannot be parsed.
    """
    time_elements_structure = self._GetValueFromStructure(
        structure, 'date_time')

    body = self._GetValueFromStructure(structure, 'body')
    reporter = self._GetValueFromStructure(structure, 'reporter')

    if key == 'rsyslog_protocol_23_line':
      priority = self._GetValueFromStructure(structure, 'priority')

      facility = self._PriorityToFacility(priority)
      message_identifier = self._GetValueFromStructure(
          structure, 'message_identifier')
      severity = self._PriorityToSeverity(priority)
    else:
      facility = None
      message_identifier = None
      severity = self._GetValueFromStructure(structure, 'severity')

    event_data = None
    if reporter == 'CRON':
      event_data = self._ParseCronMessageBody(body)
    elif reporter == 'sshd':
      event_data = self._ParseSshdMessageBody(body)

    if not event_data:
      event_data = SyslogLineEventData()

    event_data.body = body
    event_data.facility = facility
    event_data.hostname = self._GetValueFromStructure(structure, 'hostname')
    event_data.last_written_time = self._ParseTimeElements(
        time_elements_structure)
    event_data.message_identifier = message_identifier
    event_data.pid = self._GetValueFromStructure(structure, 'pid')
    event_data.reporter = reporter
    event_data.severity = severity

    parser_mediator.ProduceEventData(event_data)

  def _ParseTimeElements(self, time_elements_structure):
    """Parses date and time elements of a log line.

    Args:
      time_elements_structure (pyparsing.ParseResults): date and time elements
          of a log line.

    Returns:
      dfdatetime.TimeElements: date and time value.

    Raises:
      ParseError: if a valid date and time value cannot be derived from
          the time elements.
    """
    try:
      time_zone_minutes = 0

      if len(time_elements_structure) == 9:
        (year, month, day_of_month, hours, minutes, seconds, microseconds,
         time_zone_sign, time_zone_hours) = time_elements_structure

      else:
        (year, month, day_of_month, hours, minutes, seconds, microseconds,
         time_zone_sign, time_zone_hours, time_zone_minutes) = (
            time_elements_structure)

      time_zone_offset = (time_zone_hours * 60) + time_zone_minutes
      if time_zone_sign == '-':
        time_zone_offset *= -1

      time_elements_tuple = (
          year, month, day_of_month, hours, minutes, seconds, microseconds)

      date_time = dfdatetime_time_elements.TimeElementsInMicroseconds(
          time_elements_tuple=time_elements_tuple,
          time_zone_offset=time_zone_offset)

      return date_time

    except (IndexError, TypeError, ValueError) as exception:
      raise errors.ParseError(
          f'Unable to parse time elements with error: {exception!s}')

  def _PriorityToSeverity(self, priority):
    """Converts a syslog protocol 23 priority value to severity.

    Severity is derived from the 3 least significant bits of the priority
    value.

    Args:
      priority (int): a syslog protocol 23 priority value.

    Returns:
      str: the value from _SYSLOG_SEVERITY corresponding to severity value.
    """
    return self._SYSLOG_SEVERITY[priority & 0x07]

  def _PriorityToFacility(self, priority: int) -> str:
    """Converts a syslog protocol 23 or RFC3164 priority to facility.

    Facility is derived from the 5 most significant bits of the priority
    value.

    Args:
      priority (int): a syslog rfc3164 or protocol 23 priority value.

    Returns:
      str: the value from _SYSLOG_FACILITY corresponding to facility value.
    """
    return self._SYSLOG_FACILITY[priority >> 3]

  def CheckRequiredFormat(self, parser_mediator, text_reader):
    """Check if the log record has the minimal structure required by the parser.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      text_reader (EncodedTextReader): text reader.

    Returns:
      bool: True if this is the correct plugin, False otherwise.
    """
    try:
      structure = self._VerifyString(text_reader.lines)
    except errors.ParseError:
      return False

    time_elements_structure = self._GetValueFromStructure(
        structure, 'date_time')

    try:
      self._ParseTimeElements(time_elements_structure)
    except errors.ParseError:
      return False

    return True


class TraditionalSyslogTextPlugin(
    BaseSyslogTextPlugin, dateless_helper.DateLessLogFormatHelper):
  """Text parser plugin for traditional syslog log files."""

  NAME = 'syslog_traditional'
  DATA_FORMAT = 'Traditional system log (syslog) file'

  ENCODING = 'utf-8'

  # The reporter and facility fields can contain any printable character, but
  # to allow for processing of syslog formats that delimit the reporter and
  # facility with printable characters, we remove certain common delimiters
  # from the set of printable characters.

  _REPORTER_CHARACTERS = ''.join(
      [c for c in pyparsing.printables if c not in [':', '[', '<']])

  _FACILITY_CHARACTERS = ''.join(
      [c for c in pyparsing.printables if c not in [':', '>']])

  _SYSLOG_SEVERITY = [
      'ALERT', 'CRIT', 'DEBUG', 'EMERG', 'ERR', 'INFO', 'NOTICE', 'WARNING']

  # TODO: change pattern to allow only spaces as a field separator.
  _BODY_PATTERN = (
      r'.*?(?=($|\n\w{3}\s+\d{1,2}\s\d{2}:\d{2}:\d{2})|'
      r'($|\n\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\.\d{6}[\+|-]\d{2}:\d{2}\s)|'
      r'($|\n<\d{1,3}>1\s\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\.\d{6}[\+|-]\d{2}'
      r':\d{2}\s))')

  _ONE_OR_TWO_DIGITS = pyparsing.Word(pyparsing.nums, max=2).set_parse_action(
      lambda tokens: int(tokens[0], 10))

  _TWO_DIGITS = pyparsing.Word(pyparsing.nums, exact=2).set_parse_action(
      lambda tokens: int(tokens[0], 10))

  _THREE_LETTERS = pyparsing.Word(pyparsing.alphas, exact=3)

  _DATE_TIME = (
      _THREE_LETTERS + _ONE_OR_TWO_DIGITS +
      _TWO_DIGITS + pyparsing.Suppress(':') +
      _TWO_DIGITS + pyparsing.Suppress(':') +
      _TWO_DIGITS + pyparsing.Optional(
          pyparsing.Suppress('.') +
          pyparsing.Word(pyparsing.nums)))

  _PROCESS_IDENTIFIER = pyparsing.Word(pyparsing.nums, max=5).set_parse_action(
      lambda tokens: int(tokens[0], 10))

  _REPORTER = pyparsing.Word(_REPORTER_CHARACTERS)

  _END_OF_LINE = pyparsing.Suppress(pyparsing.LineEnd())

  # The rsyslog traditional file format (RSYSLOG_TraditionalFileFormat)
  # consists of:
  # %TIMESTAMP% %HOSTNAME% %syslogtag%%msg%
  #
  # Where %TIMESTAMP% is in year-less ctime date time format e.g.
  # Jan 22 07:54:32

  _RSYSLOG_BODY = (
      pyparsing.Word(pyparsing.printables).set_results_name('hostname') +
      _REPORTER.set_results_name('reporter') +
      pyparsing.Optional(
          pyparsing.Suppress('[') + _PROCESS_IDENTIFIER.set_results_name(
              'pid') +
          pyparsing.Suppress(']')) +
      pyparsing.Optional(
          pyparsing.Suppress('<') +
          pyparsing.Word(_FACILITY_CHARACTERS).set_results_name('facility') +
          pyparsing.Suppress('>')) +
      pyparsing.Optional(pyparsing.Suppress(':')) +
      pyparsing.Regex(_BODY_PATTERN, re.DOTALL).set_results_name('body'))

  _SYSLOG_COMMENT_END = pyparsing.Suppress('---') + _END_OF_LINE

  _SYSLOG_COMMENT_BODY = (
      pyparsing.Suppress(': ---') +
      pyparsing.SkipTo(_SYSLOG_COMMENT_END).set_results_name('body') +
      pyparsing.Suppress('---'))

  _KERNEL_SYSLOG_BODY = (
      pyparsing.Literal('kernel').set_results_name('reporter') +
      pyparsing.Suppress(':') +
      pyparsing.Regex(_BODY_PATTERN, re.DOTALL).set_results_name('body'))

  _LOG_LINE = (
      _DATE_TIME.set_results_name('date_time') + (
          _KERNEL_SYSLOG_BODY ^ _RSYSLOG_BODY ^ _SYSLOG_COMMENT_BODY) +
      _END_OF_LINE)

  _LINE_STRUCTURES = [('log_line', _LOG_LINE)]

  # Using a regular expression here is faster on non-match than the log line
  # grammar.
  VERIFICATION_GRAMMAR = pyparsing.Regex(
      r'(?P<date_time>(Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec) '
      r'( [1-9]|[1-9][0-9]) [0-9]{2}:[0-9]{2}:[0-9]{2}) \S+ .*\n')

  def _ParseRecord(self, parser_mediator, key, structure):
    """Parses a pyparsing structure.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      key (str): name of the parsed structure.
      structure (pyparsing.ParseResults): tokens from a parsed log line.

    Raises:
      ParseError: if the structure cannot be parsed.
    """
    time_elements_structure = self._GetValueFromStructure(
        structure, 'date_time')

    body = self._GetValueFromStructure(structure, 'body')
    reporter = self._GetValueFromStructure(structure, 'reporter')

    event_data = None
    if reporter == 'CRON':
      event_data = self._ParseCronMessageBody(body)
    elif reporter == 'sshd':
      event_data = self._ParseSshdMessageBody(body)

    if not event_data:
      event_data = SyslogLineEventData()

    event_data.body = body
    event_data.hostname = self._GetValueFromStructure(structure, 'hostname')
    event_data.last_written_time = self._ParseTimeElements(
        time_elements_structure)
    event_data.pid = self._GetValueFromStructure(structure, 'pid')
    event_data.reporter = reporter
    event_data.severity = self._GetValueFromStructure(structure, 'severity')

    parser_mediator.ProduceEventData(event_data)

  def _ParseTimeElements(self, time_elements_structure):
    """Parses date and time elements of a log line.

    Args:
      time_elements_structure (pyparsing.ParseResults): date and time elements
          of a log line.

    Returns:
      dfdatetime.TimeElements: date and time value.

    Raises:
      ParseError: if a valid date and time value cannot be derived from
          the time elements.
    """
    try:
      if len(time_elements_structure) == 5:
        month_string, day_of_month, hours, minutes, seconds = (
            time_elements_structure)

      else:
        # TODO: add support for fractional seconds.
        month_string, day_of_month, hours, minutes, seconds, _ = (
            time_elements_structure)

      month = self._GetMonthFromString(month_string)

      self._UpdateYear(month)

      year = self._GetRelativeYear()

      time_elements_tuple = (
          year, month, day_of_month, hours, minutes, seconds)

      date_time = dfdatetime_time_elements.TimeElements(
          is_delta=True, time_elements_tuple=time_elements_tuple)

      date_time.is_local_time = True

      return date_time

    except (IndexError, TypeError, ValueError) as exception:
      raise errors.ParseError(
          f'Unable to parse time elements with error: {exception!s}')

  def CheckRequiredFormat(self, parser_mediator, text_reader):
    """Check if the log record has the minimal structure required by the parser.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      text_reader (EncodedTextReader): text reader.

    Returns:
      bool: True if this is the correct plugin, False otherwise.
    """
    try:
      structure = self._VerifyString(text_reader.lines)
    except errors.ParseError:
      return False

    date_time_structure = self._GetValueFromStructure(structure, 'date_time')

    try:
      time_elements_structure = self._DATE_TIME.parse_string(
          date_time_structure)
    except pyparsing.ParseException:
      return False

    self._SetEstimatedYear(parser_mediator)

    try:
      self._ParseTimeElements(time_elements_structure)
    except errors.ParseError:
      return False

    return True


text_parser.TextLogParser.RegisterPlugins([
    SyslogTextPlugin, TraditionalSyslogTextPlugin])



"""Text parser plugin for iOS sysdiag log files."""

import pyparsing

from dfdatetime import time_elements as dfdatetime_time_elements

from plaso.containers import events
from plaso.lib import errors
from plaso.parsers import text_parser
from plaso.parsers.text_plugins import interface


class IOSSysdiagLogEventData(events.EventData):
  """iOS sysdiagnose log event data.

  Attributes:
    body (str): body of the event line.
    originating_call (str): call that created the entry.
    process_identifier (str): process_identifier.
    severity (str): severity of the message.
    written_time (dfdatetime.DateTimeValues): date and time the log entry was
        written.
  """

  DATA_TYPE = 'ios:sysdiag_log:entry'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.body = None
    self.originating_call = None
    self.process_identifier = None
    self.severity = None
    self.written_time = None


class IOSSysdiagLogTextPlugin(interface.TextPluginWithLineContinuation):
  """Text parser plugin for iOS mobile installation log files."""

  NAME = 'ios_sysdiag_log'
  DATA_FORMAT = 'iOS sysdiag log'

  _MONTH_DICT = {
      'jan': 1,
      'feb': 2,
      'mar': 3,
      'apr': 4,
      'may': 5,
      'jun': 6,
      'jul': 7,
      'aug': 8,
      'sep': 9,
      'oct': 10,
      'nov': 11,
      'dec': 12}

  _INTEGER = pyparsing.Word(pyparsing.nums).set_parse_action(
      lambda tokens: int(tokens[0], 10))

  _ONE_OR_TWO_DIGITS = pyparsing.Word(pyparsing.nums, max=2).set_parse_action(
      lambda tokens: int(tokens[0], 10))

  _TWO_DIGITS = pyparsing.Word(pyparsing.nums, exact=2).set_parse_action(
      lambda tokens: int(tokens[0], 10))

  _FOUR_DIGITS = pyparsing.Word(pyparsing.nums, exact=4).set_parse_action(
      lambda tokens: int(tokens[0], 10))

  _THREE_LETTERS = pyparsing.Word(pyparsing.alphas, exact=3)

  # Date and time values are formatted as: Wed Aug 11 05:51:02 2021
  _DATE_TIME = (
      _THREE_LETTERS.suppress() +
      _THREE_LETTERS +
      _ONE_OR_TWO_DIGITS +
      _TWO_DIGITS + pyparsing.Suppress(':') +
      _TWO_DIGITS + pyparsing.Suppress(':') + _TWO_DIGITS +
      _FOUR_DIGITS)

  _ORIGINATING_CALL = pyparsing.Combine(
      pyparsing.Optional(
          pyparsing.Word('+-', exact=1) + pyparsing.Literal('[') +
          pyparsing.CharsNotIn(']') + pyparsing.Literal(']')) +
      pyparsing.Optional(pyparsing.CharsNotIn(':')))

  _END_OF_LINE = pyparsing.Suppress(pyparsing.LineEnd())

  _LOG_LINE_START = (
      _DATE_TIME.set_results_name('date_time') +
      pyparsing.Suppress('[') +
      _INTEGER.set_results_name('process_identifier') +
      pyparsing.Suppress(']') + pyparsing.Suppress('<') +
      pyparsing.Word(pyparsing.alphanums).set_results_name('severity') +
      pyparsing.Suppress('>') +
      pyparsing.Suppress('(') +
      pyparsing.Word(pyparsing.alphanums).set_results_name('id') +
      pyparsing.Suppress(')') +
      _ORIGINATING_CALL.set_results_name('originating_call') +
      pyparsing.Suppress(': '))

  _LOG_LINE = (
      _LOG_LINE_START + pyparsing.restOfLine().set_results_name('body') +
      _END_OF_LINE)

  _LINE_STRUCTURES = [('log_line', _LOG_LINE)]

  VERIFICATION_GRAMMAR = _LOG_LINE

  def __init__(self):
    """Initializes a text parser plugin."""
    super().__init__()
    self._body_lines = None
    self._event_data = None

  def _ParseFinalize(self, parser_mediator):
    """Finalizes parsing.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
    """
    if self._event_data:
      self._event_data.body = ' '.join(self._body_lines)
      self._body_lines = None

      parser_mediator.ProduceEventData(self._event_data)
      self._event_data = None

  def _ParseLogline(self, structure):
    """Parses a log line.

    Args:
      structure (pyparsing.ParseResults): structure of tokens derived from
          a line of a text file.
    """
    time_elements_structure = self._GetValueFromStructure(
        structure, 'date_time')

    body = self._GetValueFromStructure(structure, 'body')

    event_data = IOSSysdiagLogEventData()
    event_data.originating_call = self._GetValueFromStructure(
        structure, 'originating_call')
    event_data.process_identifier = self._GetValueFromStructure(
        structure, 'process_identifier')
    event_data.severity = self._GetValueFromStructure(structure, 'severity')
    event_data.written_time = self._ParseTimeElements(time_elements_structure)

    self._event_data = event_data
    self._body_lines = [body]

  def _ParseRecord(self, parser_mediator, key, structure):
    """Parses a pyparsing structure.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      key (str): name of the parsed structure.
      structure (pyparsing.ParseResults): tokens from a parsed log line.

    Raises:
      ParseError: if the structure cannot be parsed.
    """
    if key == '_line_continuation':
      body = structure.replace('\n', ' ').strip()
      self._body_lines.append(body)

    else:
      if self._event_data:
        self._event_data.body = ' '.join(self._body_lines)

        parser_mediator.ProduceEventData(self._event_data)

      self._ParseLogline(structure)

  def _ParseTimeElements(self, time_elements_structure):
    """Parses date and time elements of a log line.

    Args:
      time_elements_structure (pyparsing.ParseResults): date and time elements
          of a log line.

    Returns:
      dfdatetime.TimeElements: date and time value.

    Raises:
      ParseError: if a valid date and time value cannot be derived from
          the time elements.
    """
    try:
      month_string, day_of_month, hours, minutes, seconds, year = (
          time_elements_structure)

      month = self._MONTH_DICT.get(month_string.lower(), 0)

      time_elements_tuple = (year, month, day_of_month, hours, minutes, seconds)

      return dfdatetime_time_elements.TimeElements(
          time_elements_tuple=time_elements_tuple)

    except (IndexError, TypeError, ValueError) as exception:
      raise errors.ParseError(
          f'Unable to parse time elements with error: {exception!s}')

  def _ResetState(self):
    """Resets stored values."""
    self._body_lines = None
    self._event_data = None

  def CheckRequiredFormat(self, parser_mediator, text_reader):
    """Check if the log record has the minimal structure required by the parser.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      text_reader (EncodedTextReader): text reader.

    Returns:
      bool: True if this is the correct plugin, False otherwise.
    """
    try:
      structure = self._VerifyString(text_reader.lines)
    except errors.ParseError:
      return False

    time_elements_structure = self._GetValueFromStructure(
        structure, 'date_time')

    try:
      self._ParseTimeElements(time_elements_structure)
    except errors.ParseError:
      return False

    self._ResetState()

    return True


text_parser.TextLogParser.RegisterPlugin(IOSSysdiagLogTextPlugin)



"""SQLite parser plugin for iOS Instagram threads database files.

The iOS Instagram threads database file is typically stored in:
  9368974384.db
"""

import plistlib

from plaso.containers import events
from plaso.parsers import sqlite
from plaso.parsers.plist_plugins import interface as plist_interface
from plaso.parsers.sqlite_plugins import interface


class IOSInstagramMessageEventData(events.EventData):
  """Instagram message event data.

  Attributes:
    message (str): Content of the direct message or chat sent in the thread.
    query (str): SQL query that was used to obtain the event data.
    sender_identifier (str): unique identifier (or primary key) of the sender
        account.
    sent_time (dfdatetime.DateTimeValues): Date and time when the message was
        sent.
    shared_media_identifier (str): Identifier of the media shared in the thread.
    shared_media_url (str): URL to the media content shared in the thread.
    username (str): username of the sender or participant in the thread.
    video_chat_call_identifier (str): Identifier of the video call session.
    video_chat_title (str): Title or description of the video chat.
  """

  DATA_TYPE = 'ios:instagram:message'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.message = None
    self.query = None
    self.sender_identifier = None
    self.shared_media_identifier = None
    self.shared_media_url = None
    self.sent_time = None
    self.username = None
    self.video_chat_call_identifier = None
    self.video_chat_title = None


class IOSInstagramPlugin(interface.SQLitePlugin):
  """SQLite parser plugin for iOS Instagram threads database files."""

  NAME = 'instagram_ios'
  DATA_FORMAT = 'iOS Instagram threads SQLite database (9368974384.db) file'

  REQUIRED_STRUCTURE = {
      'messages': frozenset(['message_id', 'archive']),
      'threads': frozenset(['thread_id','metadata'])}

  # Note that the threads table is processed first to determine the usernames.
  QUERIES = [
      (('SELECT thread_id, metadata FROM threads'), '_ParseThreadsRow'),
      (('SELECT message_id, archive FROM messages'), '_ParseMessagesRow')]

  SCHEMAS = [{
      'inbox_metadata': (
          'CREATE TABLE inbox_metadata ( sequence_id INTEGER NOT NULL, '
          'snapshot_at_ms REAL NOT NULL, snapshot_app_version TEXT NOT NULL, '
          'cursors BLOB NOT NULL, unseen_count_from_server INTEGER, '
          'business_messaging_high_volume_rate_limit_state INTEGER NOT NULL, '
          'row_id INTEGER PRIMARY KEY AUTOINCREMENT)'),
      'messages': (
          'CREATE TABLE messages ( message_id TEXT NOT NULL, '
          'thread_id TEXT NOT NULL, archive BLOB NOT NULL, '
          'class_name TEXT NOT NULL, '
          'row_id INTEGER PRIMARY KEY AUTOINCREMENT)'),
      'mutations': (
          'CREATE TABLE mutations ( mutation_id TEXT UNIQUE NOT NULL, '
          'archive BLOB NOT NULL, row_id INTEGER PRIMARY KEY AUTOINCREMENT)'),
      'quick_reply': (
          'CREATE TABLE quick_reply ( quickReplyId TEXT, shortcut TEXT, '
          'message TEXT, row_id INTEGER PRIMARY KEY AUTOINCREMENT)'),
      'sqlite_sequence': (
          'CREATE TABLE sqlite_sequence(name,seq)'),
      'thread_client_state': (
          'CREATE TABLE thread_client_state ( thread_id TEXT NOT NULL, '
          'archive BLOB NOT NULL, row_id INTEGER PRIMARY KEY AUTOINCREMENT)'),
      'threads': (
          'CREATE TABLE threads ( thread_id TEXT UNIQUE NOT NULL, '
          'thread_id_v2 TEXT, viewer_id TEXT NOT NULL, metadata BLOB, '
          'thread_messages_range BLOB, visual_message_info BLOB, '
          'row_id INTEGER PRIMARY KEY AUTOINCREMENT)')}]

  def __init__(self):
    """Initializes the plugin."""
    super().__init__()
    self._plist_decoder = plist_interface.NSKeyedArchiverDecoder()

  def _ParseMessagesRow(
      self, parser_mediator, query, row, cache=None, **unused_kwargs):
    """Parses a row in the messages table.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      query (str): query that created the row.
      row (sqlite3.Row): row.
      cache (Optional[SQLiteCache]): cache.
    """
    query_hash = hash(query)

    message_identifier = self._GetRowValue(query_hash, row, 'message_id')

    archive = self._GetRowValue(query_hash, row, 'archive')
    if not archive:
      parser_mediator.ProduceExtractionWarning(
          f'Message: {message_identifier:s} missing archive value')
      return

    if not archive.startswith(b'bplist'):
      parser_mediator.ProduceExtractionWarning(
          f'Message: {message_identifier:s} unsupported archive value - not a '
          f'binary plist.')
      return

    try:
      top_level_object = plistlib.loads(archive)

    except plistlib.InvalidFileException as exception:
      parser_mediator.ProduceExtractionWarning(
          f'Message: {message_identifier:s} unable to parse archive plist with '
          f'error: {exception!s}')
      return

    if not self._plist_decoder.IsEncoded(top_level_object):
      parser_mediator.ProduceExtractionWarning(
          f'Message: {message_identifier:s} unsupported archive plist '
          f'encoding.')
      return

    decoded_plist = self._plist_decoder.Decode(top_level_object)

    content = decoded_plist.get('IGDirectPublishedMessageContent*content', {})
    metadata = decoded_plist.get(
        'IGDirectPublishedMessageMetadata*metadata', {})

    media = content.get('IGDirectPublishedMessageMedia*media', {})
    thread_activity = content.get(
        'IGDirectThreadActivityAnnouncement*threadActivity', {})

    photo = media.get(
        'IGDirectPublishedMessagePermanentMedia*permanentMedia', {}).get(
            'IGPhoto*photo', {})
    image_versions = photo.get('imageVersions')

    event_data = IOSInstagramMessageEventData()
    event_data.message = content.get('NSString*string')
    event_data.query = query
    event_data.sender_identifier = metadata.get('NSString*senderPk')
    event_data.sent_time = metadata.get('NSDate*serverTimestamp')
    event_data.shared_media_identifier = photo.get('kIGPhotoMediaID')
    event_data.video_chat_call_identifier = thread_activity.get(
        'NSString*videoCallId')
    event_data.video_chat_title = thread_activity.get('NSString*voipTitle')

    if event_data.sender_identifier:
      usernames = cache.GetResults('usernames', default_value={})
      event_data.username = usernames.get(event_data.sender_identifier)
    if image_versions:
      event_data.shared_media_url = image_versions[0].get("url")

    parser_mediator.ProduceEventData(event_data)

  def _ParseThreadsRow(
      self, parser_mediator, query, row, cache=None, **unused_kwargs):
    """Parses a row in the threads table.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      query (str): query that created the row.
      row (sqlite3.Row): row.
      cache (Optional[SQLiteCache]): cache.
    """
    query_hash = hash(query)

    thread_identifier = self._GetRowValue(query_hash, row, 'thread_id')

    metadata = self._GetRowValue(query_hash, row, 'metadata')
    if not metadata:
      return

    if not metadata.startswith(b'bplist'):
      parser_mediator.ProduceExtractionWarning(
          f'Thread: {thread_identifier:s} unsupported metadata - not a binary '
          f'plist.')
      return

    try:
      top_level_object = plistlib.loads(metadata)

    except plistlib.InvalidFileException as exception:
      parser_mediator.ProduceExtractionWarning(
          f'Thread: {thread_identifier:s} unable to parse metadata plist with '
          f'error: {exception!s}')
      return

    if not self._plist_decoder.IsEncoded(top_level_object):
      parser_mediator.ProduceExtractionWarning(
          f'Thread: {thread_identifier:s} unsupported metadata plist encoding.')
      return

    decoded_plist = self._plist_decoder.Decode(top_level_object)

    usernames = cache.GetResults('usernames', default_value={})

    inviter = decoded_plist.get('IGUser*inviter', {})
    user_identifier = inviter.get("pk")
    if user_identifier:
      usernames[user_identifier] = inviter.get("fullName") or None

    for user in decoded_plist.get('NSArray<IGUser *>*users') or []:
      user_identifier = user.get("pk")
      if user_identifier:
        usernames[user_identifier] = user.get("fullName") or None

    cache.SetResults('usernames', usernames)


sqlite.SQLiteParser.RegisterPlugin(IOSInstagramPlugin)



"""Report related attribute container definitions."""

from acstore.containers import interface
from acstore.containers import manager


class AnalysisReport(interface.AttributeContainer):
  """Analysis report attribute container.

  Attributes:
    analysis_counter (collections.Counter): counter of analysis results, for
         example number of events analyzed and tagged.
    event_filter (str): event filter expression that was used when the analysis
        plugin was run.
    plugin_name (str): name of the analysis plugin that generated the report.
    text (str): report text.
    time_compiled (int): timestamp of the date and time the report was compiled.
  """
  CONTAINER_TYPE = 'analysis_report'

  def __init__(self, plugin_name=None, text=None):
    """Initializes the analysis report.

    Args:
      plugin_name (Optional[str]): name of the analysis plugin that generated
          the report.
      text (Optional[str]): report text.
    """
    super().__init__()
    self.analysis_counter = None
    self.event_filter = None
    self.plugin_name = plugin_name
    # TODO: kept for backwards compatibility.
    self.text = text
    self.time_compiled = None

  def CopyToDict(self):
    """Copies the attribute container to a dictionary.

    Returns:
      dict[str, object]: attribute values per name.
    """
    dictionary = {}
    for attribute_name, attribute_value in self.GetAttributes():
      if attribute_value is None:
        continue

      dictionary[attribute_name] = attribute_value

    return dictionary


manager.AttributeContainersManager.RegisterAttributeContainer(AnalysisReport)



"""SQLite parser plugin for Android Native Downloads database files.

The Android native downloads database file is typically stored as:
  com.android.providers.downloads/databases/downloads.db

Also see :
  STATUS_* and DESTINATION_* constants:
    https://android.googlesource.com/platform/frameworks/base/+/refs/heads/master/core/java/android/provider/Downloads.java
  ERROR_*, PAUSED_*, and VISIBILITY_* constants:
    https://android.googlesource.com/platform/frameworks/base/+/HEAD/core/java/android/app/DownloadManager.java
"""

from dfdatetime import java_time as dfdatetime_java_time

from plaso.containers import events
from plaso.parsers import sqlite
from plaso.parsers.sqlite_plugins import interface


class AndroidNativeDownloadsEventData(events.EventData):
  """Android Native Downloads (DownloadProvider) event data.

  Attributes:
    current_bytes (int): Number of bytes download so far.
    deleted (bool): Set to true if this download is deleted.
       Also Removed from the database when MediaProvider database
       deletes the metadata associated with this downloaded file.
    description (str): The client-supplied description of this download. This
        will be displayed in system notifications.
    destination (int): Flag that controls download destination. Also see the
        DESTINATION_* constants for a list of supported values.
    error_msg (str): The column with errorMsg for a failed downloaded.
        Used only for debugging purposes.
    e_tag (str): ETag of this file.
    identifier (int): identifier of the download.
    is_visible_in_downloads_ui (int):  Whether or not this download should
        be displayed in the system's Downloads UI. Defaults to true.
    media_provider_uri (str): The URI to corresponding entry in MediaProvider
        for this downloaded entry. If an entry is deleted from downloaded list,
        it is also deleted from MediaProvider DB.
    mime_type (str): Internet Media Type of the downloaded file.
    modification_time (dfdatetime.DateTimeValues): Last content modification
        date and time of downloaded file.
    notification_package (str): Package name associated with notification
        of a running download.
    status (int): Holds one of the STATUS_* constants.
        If an error occurred, this holds the HTTP Status Error Code (RFC 2616),
        otherwise it holds one of the ERROR_* constants.
        If the download is paused, this holds one of the PAUSED_* constants.
    saved_to (str): Path to the downloaded file on disk.
    title (str): Title of the download.
    total_bytes (int): Total size of the download in bytes.
    ui_visibility (int): Flags that control if the download is displayed by the
        UI. Also see VISIBILITY_* constants for a list of supported values.
    uri (str): Downloaded URI.
  """

  DATA_TYPE = 'android:sqlite:downloads'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.current_bytes = None
    self.deleted = None
    self.description = None
    self.destination = None
    self.error_msg = None
    self.e_tag = None
    self.identifier = None
    self.is_visible_in_downloads_ui = None
    self.media_provider_uri = None
    self.mime_type = None
    self.modification_time = None
    self.notification_package = None
    self.saved_to = None
    self.status = None
    self.title = None
    self.total_bytes = None
    self.ui_visibility = None
    self.uri = None


class AndroidNativeDownloadsPlugin(interface.SQLitePlugin):
  """SQLite parser plugin for Android native downloads database file."""

  NAME = 'android_native_downloads'
  DATA_FORMAT = 'Android native downloads SQLite database (downloads.db) file'

  REQUIRED_STRUCTURE = {
      'downloads': frozenset([
          '_id', 'uri', '_data', 'mimetype', 'destination', 'visibility',
          'status', 'lastmod', 'notificationpackage', 'total_bytes',
          'current_bytes', 'etag', 'title', 'description',
          'is_visible_in_downloads_ui', 'mediaprovider_uri', 'deleted',
          'errorMsg'])}

  QUERIES = [(
        ('SELECT _id, uri, _data, mimetype, destination, visibility, status, '
         'lastmod, notificationpackage, total_bytes, current_bytes, etag, '
         'title, description, is_visible_in_downloads_ui, mediaprovider_uri, '
         'deleted, errorMsg FROM downloads'), 'ParseDownloadsRow')]

  SCHEMAS = [{
      'android_metadata': (
          'CREATE TABLE android_metadata (locale TEXT)'),
      'downloads': (
          'CREATE TABLE downloads(_id INTEGER PRIMARY KEY AUTOINCREMENT, '
          'uri TEXT, method INTEGER, entity TEXT, no_integrity BOOLEAN, '
          'hint TEXT, otaupdate BOOLEAN, _data TEXT, mimetype TEXT, '
          'destination INTEGER, no_system BOOLEAN, visibility INTEGER, '
          'control INTEGER, status INTEGER, numfailed INTEGER, '
          'lastmod BIGINT, notificationpackage TEXT, notificationclass TEXT, '
          'notificationextras TEXT, cookiedata TEXT, useragent TEXT, '
          'referer TEXT, total_bytes INTEGER, current_bytes INTEGER, '
          'etag TEXT, uid INTEGER, otheruid INTEGER, title TEXT, '
          'description TEXT, scanned BOOLEAN, '
          'is_public_api INTEGER NOT NULL DEFAULT 0, '
          'allow_roaming INTEGER NOT NULL DEFAULT 0, '
          'allowed_network_types INTEGER NOT NULL DEFAULT 0, '
          'is_visible_in_downloads_ui INTEGER NOT NULL DEFAULT 1, '
          'bypass_recommended_size_limit INTEGER NOT NULL DEFAULT 0, '
          'mediaprovider_uri TEXT, deleted BOOLEAN NOT NULL DEFAULT 0, '
          'errorMsg TEXT, allow_metered INTEGER NOT NULL DEFAULT 1, '
          'allow_write BOOLEAN NOT NULL DEFAULT 0, '
          'flags INTEGER NOT NULL DEFAULT 0, '
          'mediastore_uri TEXT DEFAULT NULL)'),
      'request_headers': (
          'CREATE TABLE request_headers(id INTEGER PRIMARY KEY AUTOINCREMENT,'
          'download_id INTEGER NOT NULL, '
          'header TEXT NOT NULL,value TEXT NOT NULL)'),
      'sqlite_sequence': (
          'CREATE TABLE sqlite_sequence(name,seq)')}]

  def _GetDateTimeRowValue(self, query_hash, row, value_name):
    """Retrieves a date and time value from the row.

    Args:
      query_hash (int): hash of the query, that uniquely identifies the query
          that produced the row.
      row (sqlite3.Row): row.
      value_name (str): name of the value.

    Returns:
      dfdatetime.JavaTime: date and time value or None if not available.
    """
    timestamp = self._GetRowValue(query_hash, row, value_name)
    if timestamp is None:
      return None

    return dfdatetime_java_time.JavaTime(timestamp=timestamp)

  def ParseDownloadsRow(self, parser_mediator, query, row, **unused_kwargs):
    """Parses a download row.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      query (str): query that created the row.
      row (sqlite3.Row): row.
    """
    query_hash = hash(query)

    event_data = AndroidNativeDownloadsEventData()
    event_data.current_bytes = self._GetRowValue(
        query_hash, row, 'current_bytes')
    event_data.deleted = self._GetRowValue(query_hash, row, 'deleted')
    # Description defaults to an empty string.
    event_data.description = self._GetRowValue(
        query_hash, row, 'description') or None
    event_data.destination = self._GetRowValue(query_hash, row, 'destination')
    event_data.error_msg = self._GetRowValue(query_hash, row, 'errorMsg')
    event_data.e_tag = self._GetRowValue(query_hash, row, 'etag')
    event_data.identifier = self._GetRowValue(query_hash, row, '_id')
    event_data.is_visible_in_downloads_ui = self._GetRowValue(
        query_hash, row, 'is_visible_in_downloads_ui')
    event_data.modification_time = self._GetDateTimeRowValue(
        query_hash, row, 'lastmod')
    event_data.media_provider_uri = self._GetRowValue(
        query_hash, row, 'mediaprovider_uri')
    event_data.mime_type = self._GetRowValue(query_hash, row, 'mimetype')
    event_data.notification_package = self._GetRowValue(
        query_hash, row, 'notificationpackage')
    event_data.saved_to = self._GetRowValue(query_hash, row, '_data')
    event_data.status = self._GetRowValue(query_hash, row, 'status')
    event_data.title = self._GetRowValue(query_hash, row, 'title')
    event_data.total_bytes = self._GetRowValue(query_hash, row, 'total_bytes')
    event_data.ui_visibility = self._GetRowValue(query_hash, row, 'visibility')
    event_data.uri = self._GetRowValue(query_hash, row, 'uri')

    parser_mediator.ProduceEventData(event_data)


sqlite.SQLiteParser.RegisterPlugin(AndroidNativeDownloadsPlugin)



"""Plist parser plugin for Safari history plist files."""

from dfdatetime import cocoa_time as dfdatetime_cocoa_time

from plaso.containers import events
from plaso.parsers import plist
from plaso.parsers.plist_plugins import interface


class SafariHistoryEventData(events.EventData):
  """Safari history event data.

  Attributes:
    display_title (str): display title of the webpage visited.
    last_visited_time (dfdatetime.DateTimeValues): date and time the URL was
        last visited.
    title (str): title of the webpage visited.
    url (str): URL visited.
    visit_count (int): number of times the website was visited.
    was_http_non_get (bool): True if the webpage was visited using a non-GET
        HTTP request.
  """

  DATA_TYPE = 'safari:history:visit'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.display_title = None
    self.last_visited_time = None
    self.title = None
    self.url = None
    self.visit_count = None
    self.was_http_non_get = None


class SafariHistoryPlugin(interface.PlistPlugin):
  """Plist parser plugin for Safari history plist files."""

  NAME = 'safari_history'
  DATA_FORMAT = 'Safari history plist file'

  PLIST_PATH_FILTERS = frozenset([
      interface.PlistPathFilter('History.plist')])

  PLIST_KEYS = frozenset(['WebHistoryDates', 'WebHistoryFileVersion'])

  def _GetDateTimeValueFromTimestamp(
      self, parser_mediator, plist_key, plist_value_name):
    """Retrieves a date and time value from a Cocoa timestamp.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      plist_key (object): plist key.
      plist_value_name (str): name of the value in the plist key.

    Returns:
      dfdatetime.TimeElements: date and time or None if not available.
    """
    timestamp_string = plist_key.get(plist_value_name)
    if not timestamp_string:
      return None

    try:
      timestamp = float(timestamp_string)
    except (TypeError, ValueError):
      parser_mediator.ProduceExtractionWarning((
          f'unable to convert Cocoa timestamp: '
          f'{timestamp_string:s} to a floating-point value'))
      return None

    return dfdatetime_cocoa_time.CocoaTime(timestamp=timestamp)

  # pylint: disable=arguments-differ
  def _ParsePlist(self, parser_mediator, match=None, **unused_kwargs):
    """Extracts Safari history items.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      match (Optional[dict[str: object]]): keys extracted from PLIST_KEYS.
    """
    format_version = match.get('WebHistoryFileVersion')
    if format_version != 1:
      parser_mediator.ProduceExtractionWarning(
          f'unsupported Safari history version: {format_version!s}')
      return

    for history_entry in match.get('WebHistoryDates', {}):
      display_title = history_entry.get('displayTitle')
      title = history_entry.get('title')

      event_data = SafariHistoryEventData()
      event_data.last_visited_time = self._GetDateTimeValueFromTimestamp(
          parser_mediator, history_entry, 'lastVisitedDate')
      event_data.title = title
      event_data.url = history_entry.get('')
      event_data.visit_count = history_entry.get('visitCount')
      event_data.was_http_non_get = history_entry.get(
          'lastVisitWasHTTPNonGet', None)

      if display_title != title:
        event_data.display_title = display_title

      parser_mediator.ProduceEventData(event_data)


plist.PlistParser.RegisterPlugin(SafariHistoryPlugin)



"""The default event formatter."""

from acstore.containers import interface as containers_interface

from dfdatetime import interface as dfdatetime_interface

from plaso.formatters import interface


class DefaultEventFormatter(interface.BasicEventFormatter):
  """Formatter for events that do not have any defined formatter."""

  DATA_TYPE = 'event'
  FORMAT_STRING = '<WARNING DEFAULT FORMATTER> Attributes: {attribute_values}'
  FORMAT_STRING_SHORT = '<DEFAULT> {attribute_values}'

  _RESERVED_VARIABLE_NAMES = frozenset([
      '_event_values_hash',
      '_parser_chain',
      'data_type',
      'date_time',
      'path_spec',
      'timestamp',
      'timestamp_desc'])

  def __init__(self):
    """Initializes a default event formatter."""
    super().__init__(
        data_type=self.DATA_TYPE, format_string=self.FORMAT_STRING,
        format_string_short=self.FORMAT_STRING_SHORT)

  def _FormatMessage(self, format_string, event_values):
    """Determines the formatted message.

    Args:
      format_string (str): message format string.
      event_values (dict[str, object]): event values.

    Returns:
      str: formatted message.
    """
    text_pieces = []
    for name, value in event_values.items():
      # Ignore reserved variable names.
      if name in self._RESERVED_VARIABLE_NAMES:
        continue

      # Ignore attribute container identifier values.
      if isinstance(value, containers_interface.AttributeContainerIdentifier):
        continue

      # Ignore date and time values.
      if isinstance(value, dfdatetime_interface.DateTimeValues):
        continue

      if (isinstance(value, list) and value and
          isinstance(value[0], dfdatetime_interface.DateTimeValues)):
        continue

      text_pieces.append(f'{name:s}: {value!s}')

    return super()._FormatMessage(
        format_string, {'attribute_values': ' '.join(text_pieces)})



"""Bencode parser plugin interface."""

import abc

from plaso.parsers import plugins


class BencodePlugin(plugins.BasePlugin):
  """Bencode parser plugin interface."""

  NAME = 'bencode_plugin'
  DATA_FORMAT = 'Bencoded file'

  # _BENCODE_KEYS is a list of keys required by a plugin.
  # This is expected to be overridden by the processing plugin.
  # Ex. frozenset(['activity-date', 'done-date'])
  _BENCODE_KEYS = frozenset(['any'])

  def CheckRequiredKeys(self, bencode_file):
    """Checks if the bencode file has the minimal keys required by the plugin.

    Args:
      bencode_file (BencodeFile): bencode file.

    Returns:
      bool: True if the bencode file has the minimum keys defined by the plugin,
          or False if it does not or no required keys are defined. The bencode
          file can have more keys than specified by the plugin and still return
          True.
    """
    if not self._BENCODE_KEYS:
      return False

    return bencode_file.keys.issuperset(self._BENCODE_KEYS)

  # pylint: disable=arguments-differ
  @abc.abstractmethod
  def Process(self, parser_mediator, bencode_file=None, **kwargs):
    """Extracts events from a bencode file.

    This is the main method that a Bencode plugin needs to implement.

    The contents of the bencode keys defined in _BENCODE_KEYS can be made
    available to the plugin as both a matched{'KEY': 'value'} and as the
    entire bencoded data dictionary.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      bencode_file (Optional[BencodeFile]): bencode file.
    """



"""SQLite parser plugin for Android WebviewCache database files."""

from dfdatetime import java_time as dfdatetime_java_time

from plaso.containers import events
from plaso.parsers import sqlite
from plaso.parsers.sqlite_plugins import interface


class AndroidWebViewCacheEventData(events.EventData):
  """Android WebViewCache event data.

  Attributes:
    content_length (int): size of the cached content.
    expiration_time (dfdatetime.DateTimeValues): date and time the cache
        entry expires.
    last_modified_time (dfdatetime.DateTimeValues): date and time the cache
        entry was last modified.
    query (str): SQL query that was used to obtain the event data.
    url (str): URL the content was retrieved from.
  """

  DATA_TYPE = 'android:webviewcache'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.content_length = None
    self.expiration_time = None
    self.last_modified_time = None
    self.query = None
    self.url = None


class AndroidWebViewCachePlugin(interface.SQLitePlugin):
  """SQLite parser plugin for Android WebviewCache database files."""

  NAME = 'android_webviewcache'
  DATA_FORMAT = 'Android WebViewCache SQLite database file'

  REQUIRED_STRUCTURE = {
      'android_metadata': frozenset([]),
      'cache': frozenset(['url', 'contentlength', 'expires', 'lastmodify'])}

  QUERIES = frozenset([
      ('SELECT url, contentlength, expires, lastmodify FROM cache',
       'ParseRow')])

  SCHEMAS = [{
      'android_metadata': (
          'CREATE TABLE android_metadata (locale TEXT)'),
      'cache': (
          'CREATE TABLE cache (_id INTEGER PRIMARY KEY, url TEXT, filepath '
          'TEXT, lastmodify TEXT, etag TEXT, expires INTEGER, expiresstring '
          'TEXT, mimetype TEXT, encoding TEXT, httpstatus INTEGER, location '
          'TEXT, contentlength INTEGER, contentdisposition TEXT, UNIQUE (url) '
          'ON CONFLICT REPLACE)')}]

  def _GetDateTimeRowValue(self, query_hash, row, value_name):
    """Retrieves a date and time value from the row.

    Args:
      query_hash (int): hash of the query, that uniquely identifies the query
          that produced the row.
      row (sqlite3.Row): row.
      value_name (str): name of the value.

    Returns:
      dfdatetime.JavaTime: date and time value or None if not available.
    """
    timestamp = self._GetRowValue(query_hash, row, value_name)
    if timestamp is None:
      return None

    return dfdatetime_java_time.JavaTime(timestamp=timestamp)

  def ParseRow(self, parser_mediator, query, row, **unused_kwargs):
    """Parses a row from the database.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      query (str): query that created the row.
      row (sqlite3.Row): row.
    """
    query_hash = hash(query)

    event_data = AndroidWebViewCacheEventData()
    event_data.content_length = self._GetRowValue(
        query_hash, row, 'contentlength')
    event_data.expiration_time = self._GetDateTimeRowValue(
        query_hash, row, 'expires')
    event_data.last_modified_time = self._GetDateTimeRowValue(
        query_hash, row, 'lastmodify')
    event_data.query = query
    event_data.url = self._GetRowValue(query_hash, row, 'url')

    parser_mediator.ProduceEventData(event_data)


sqlite.SQLiteParser.RegisterPlugin(AndroidWebViewCachePlugin)



"""The log2timeline CLI tool."""

import argparse
import sys
import textwrap

import plaso

# The following import makes sure the output modules are registered.
from plaso import output  # pylint: disable=unused-import

from plaso.analyzers.hashers import manager as hashers_manager
from plaso.cli import extraction_tool
from plaso.cli import views
from plaso.cli.helpers import manager as helpers_manager
from plaso.lib import definitions
from plaso.lib import errors
from plaso.lib import loggers
from plaso.parsers import manager as parsers_manager


class Log2TimelineTool(extraction_tool.ExtractionTool):
  """Log2timeline CLI tool.

  Attributes:
    dependencies_check (bool): True if the availability and versions of
        dependencies should be checked.
    list_archive_types (bool): True if the archive types should be listed.
    list_hashers (bool): True if the hashers should be listed.
    list_parsers_and_plugins (bool): True if the parsers and plugins should
        be listed.
    list_profilers (bool): True if the profilers should be listed.
    show_info (bool): True if information about hashers, parsers, plugins,
        etc. should be shown.
  """

  NAME = 'log2timeline'
  DESCRIPTION = textwrap.dedent('\n'.join([
      '',
      ('log2timeline is a command line tool to extract events from '
       'individual '),
      'files, recursing a directory (e.g. mount point) or storage media ',
      'image or device.',
      '',
      'More information can be gathered from here:',
      '    https://plaso.readthedocs.io/en/latest/sources/user/'
      'Using-log2timeline.html',
      '']))

  EPILOG = textwrap.dedent('\n'.join([
      '',
      'Example usage:',
      '',
      'Run the tool against a storage media image (full kitchen sink)',
      '    log2timeline.py /cases/mycase/storage.plaso ímynd.dd',
      '',
      'Instead of answering questions, indicate some of the options on the',
      'command line (including data from particular VSS stores).',
      '    log2timeline.py --vss_stores 1,2 /cases/plaso_vss.plaso image.E01',
      '',
      'And that is how you build a timeline using log2timeline...',
      '']))

  def __init__(self, input_reader=None, output_writer=None):
    """Initializes a log2timeline CLI tool.

    Args:
      input_reader (Optional[InputReader]): input reader, where None indicates
          that the stdin input reader should be used.
      output_writer (Optional[OutputWriter]): output writer, where None
          indicates that the stdout output writer should be used.
    """
    super().__init__(
        input_reader=input_reader, output_writer=output_writer)
    self._storage_serializer_format = definitions.SERIALIZER_FORMAT_JSON

    self.dependencies_check = True
    self.list_archive_types = False
    self.list_hashers = False
    self.list_parsers_and_plugins = False
    self.list_profilers = False
    self.show_info = False

  def _GetPluginData(self):
    """Retrieves the version and various plugin information.

    Returns:
      dict[str, list[str]]: available parsers and plugins.
    """
    return_dict = {}

    return_dict['Versions'] = [
        ('plaso engine', plaso.__version__),
        ('python', sys.version)]

    hashers_information = hashers_manager.HashersManager.GetHashersInformation()
    parsers_information = parsers_manager.ParsersManager.GetParsersInformation()
    plugins_information = (
        parsers_manager.ParsersManager.GetParserPluginsInformation())
    presets_information = self._presets_manager.GetPresetsInformation()

    return_dict['Hashers'] = hashers_information
    return_dict['Parsers'] = parsers_information
    return_dict['Parser Plugins'] = plugins_information
    return_dict['Parser Presets'] = presets_information

    return return_dict

  def AddStorageOptions(self, argument_group):  # pylint: disable=arguments-renamed
    """Adds the storage options to the argument group.

    Args:
      argument_group (argparse._ArgumentGroup): argparse argument group.
    """
    argument_group.add_argument(
        '--storage_file', '--storage-file', dest='storage_file', metavar='PATH',
        type=str, default=None, help=(
            'The path of the storage file. If not specified, one will be made '
            'in the form <timestamp>-<source>.plaso'))

  def ParseArguments(self, arguments):
    """Parses the command line arguments.

    Args:
      arguments (list[str]): command line arguments.

    Returns:
      bool: True if the arguments were successfully parsed.
    """
    loggers.ConfigureLogging()

    argument_parser = argparse.ArgumentParser(
        description=self.DESCRIPTION, epilog=self.EPILOG, add_help=False,
        formatter_class=argparse.RawDescriptionHelpFormatter)

    self.AddBasicOptions(argument_parser)

    data_location_group = argument_parser.add_argument_group(
        'data location arguments')

    argument_helper_names = ['artifact_definitions', 'data_location']
    helpers_manager.ArgumentHelperManager.AddCommandLineArguments(
        data_location_group, names=argument_helper_names)

    extraction_group = argument_parser.add_argument_group(
        'extraction arguments')

    argument_helper_names = [
        'archives', 'artifact_filters', 'extraction', 'filter_file', 'hashers',
        'parsers', 'yara_rules']
    helpers_manager.ArgumentHelperManager.AddCommandLineArguments(
        extraction_group, names=argument_helper_names)

    self.AddStorageMediaImageOptions(extraction_group)
    self.AddExtractionOptions(extraction_group)
    self.AddVSSProcessingOptions(extraction_group)
    self.AddCredentialOptions(extraction_group)

    info_group = argument_parser.add_argument_group('informational arguments')

    self.AddInformationalOptions(info_group)

    info_group.add_argument(
        '--info', dest='show_info', action='store_true', default=False,
        help='Print out information about supported plugins and parsers.')

    info_group.add_argument(
        '--use_markdown', '--use-markdown', dest='use_markdown',
        action='store_true', default=False, help=(
            'Output lists in Markdown format use in combination with '
            '"--hashers list", "--parsers list" or "--timezone list"'))

    info_group.add_argument(
        '--no_dependencies_check', '--no-dependencies-check',
        dest='dependencies_check', action='store_false', default=True,
        help='Disable the dependencies check.')

    self.AddLogFileOptions(info_group)

    helpers_manager.ArgumentHelperManager.AddCommandLineArguments(
        info_group, names=['status_view'])

    processing_group = argument_parser.add_argument_group(
        'processing arguments')

    self.AddPerformanceOptions(processing_group)
    self.AddProcessingOptions(processing_group)

    processing_group.add_argument(
        '--sigsegv_handler', '--sigsegv-handler', dest='sigsegv_handler',
        action='store_true', default=False, help=(
            'Enables the SIGSEGV handler. WARNING this functionality is '
            'experimental and will a deadlock worker process if a real '
            'segfault is caught, but not signal SIGSEGV. This functionality '
            'is therefore primarily intended for debugging purposes'))

    profiling_group = argument_parser.add_argument_group('profiling arguments')

    helpers_manager.ArgumentHelperManager.AddCommandLineArguments(
        profiling_group, names=['profiling'])

    storage_group = argument_parser.add_argument_group('storage arguments')

    self.AddStorageOptions(storage_group)

    helpers_manager.ArgumentHelperManager.AddCommandLineArguments(
        storage_group, names=['storage_format'])

    argument_parser.add_argument(
        self._SOURCE_OPTION, action='store', metavar='SOURCE', nargs='?',
        default=None, type=str, help=(
            'Path to a source device, file or directory. If the source is '
            'a supported storage media device or image file, archive file '
            'or a directory, the files within are processed recursively.'))

    try:
      options = argument_parser.parse_args(arguments)
    except UnicodeEncodeError:
      # If we get here we are attempting to print help in a non-Unicode
      # terminal.
      self._output_writer.Write('\n')
      self._output_writer.Write(argument_parser.format_help())
      return False

    # Properly prepare the attributes according to local encoding.
    if self.preferred_encoding == 'ascii':
      self._PrintUserWarning((
          'the preferred encoding of your system is ASCII, which is not '
          'optimal for the typically non-ASCII characters that need to be '
          'parsed and processed. This will most likely result in an error.'))

    try:
      self.ParseOptions(options)
    except errors.BadConfigOption as exception:
      self._output_writer.Write(f'ERROR: {exception!s}\n')
      self._output_writer.Write('\n')
      self._output_writer.Write(argument_parser.format_usage())
      return False

    self._command_line_arguments = self.GetCommandLineArguments()

    self._WaitUserWarning()

    loggers.ConfigureLogging(
        debug_output=self._debug_mode, filename=self._log_file,
        quiet_mode=self._quiet_mode)

    return True

  def ParseOptions(self, options):
    """Parses the options.

    Args:
      options (argparse.Namespace): command line arguments.

    Raises:
      BadConfigOption: if the options are invalid.
    """
    # The extraction options are dependent on the data location.
    helpers_manager.ArgumentHelperManager.ParseOptions(
        options, self, names=['data_location'])

    self._ReadParserPresetsFromFile()

    # Check the list options first otherwise required options will raise.
    argument_helper_names = ['archives', 'hashers', 'parsers', 'profiling']
    helpers_manager.ArgumentHelperManager.ParseOptions(
        options, self, names=argument_helper_names)

    self._ParseExtractionOptions(options)

    self.list_archive_types = self._archive_types_string == 'list'
    self.list_hashers = self._hasher_names_string == 'list'
    self.list_parsers_and_plugins = self._parser_filter_expression == 'list'
    self.list_profilers = self._profilers == 'list'

    self.show_info = getattr(options, 'show_info', False)
    self.show_troubleshooting = getattr(options, 'show_troubleshooting', False)

    if getattr(options, 'use_markdown', False):
      self._views_format_type = views.ViewsFactory.FORMAT_TYPE_MARKDOWN

    self.dependencies_check = getattr(options, 'dependencies_check', True)

    if (self.list_archive_types or self.list_hashers or
        self.list_language_tags or self.list_parsers_and_plugins or
        self.list_profilers or self.list_time_zones or self.show_info or
        self.show_troubleshooting):
      return

    self._ParseInformationalOptions(options)

    argument_helper_names = [
        'artifact_definitions', 'artifact_filters', 'extraction',
        'filter_file', 'status_view', 'storage_format', 'yara_rules']
    helpers_manager.ArgumentHelperManager.ParseOptions(
        options, self, names=argument_helper_names)

    self._ParseLogFileOptions(options)

    self._ParseStorageMediaOptions(options)

    self._ParsePerformanceOptions(options)
    self._ParseProcessingOptions(options)

    self._storage_file_path = self.ParseStringOption(options, 'storage_file')
    if not self._storage_file_path:
      self._storage_file_path = self._GenerateStorageFileName()

    if not self._storage_file_path:
      raise errors.BadConfigOption('Missing storage file option.')

    serializer_format = getattr(
        options, 'serializer_format', definitions.SERIALIZER_FORMAT_JSON)
    if serializer_format not in definitions.SERIALIZER_FORMATS:
      raise errors.BadConfigOption(
          f'Unsupported storage serializer format: {serializer_format:s}')
    self._storage_serializer_format = serializer_format

    helpers_manager.ArgumentHelperManager.ParseOptions(
        options, self, names=['status_view'])

    self._enable_sigsegv_handler = getattr(options, 'sigsegv_handler', False)

    self._EnforceProcessMemoryLimit(self._process_memory_limit)

  def ShowInfo(self):
    """Shows information about available hashers, parsers, plugins, etc."""
    title = ' log2timeline/plaso information '
    self._output_writer.Write(f'{title:=^80s}\n')

    plugin_list = self._GetPluginData()
    for header, data in plugin_list.items():
      table_view = views.ViewsFactory.GetTableView(
          self._views_format_type, column_names=['Name', 'Description'],
          title=header)
      for entry_header, entry_data in sorted(data):
        table_view.AddRow([entry_header, entry_data])
      table_view.Write(self._output_writer)
